#ifndef NETWORK_H
#define NETWORK_H

#include <time.h>
#include <stdint.h>

typedef struct
 {
  int sock;
  char address[17];
  time_t birth, last_reception;
  int complete;
  uint32_t sent, recieved;
  
  uint8_t *send_start, *send_end, *send_alloc,
          *recv_start, *recv_end, *recv_alloc;
  
  void * custom_data;
 } Connection;

int networkInit(void);
void networkUninit(void);

int networkServerInit(uint16_t port);
void networkServerUninit(void);
void networkUpdateServer(int(*connect)(Connection *, void **), int(*update)(Connection *, void **), void(*disconnect)(Connection *, void **));

uint16_t networkClients(void);
Connection *networkClient(uint16_t);

int networkConnectToServer(Connection *, const char *, unsigned short);

int networkSendBuffer(Connection *, const uint8_t *start, const uint8_t *end);
const uint8_t *networkGetReadBuffer(Connection *);
uint16_t networkWaitingBytes(Connection *);
void networkSkipWaitingBytes(Connection *, uint16_t);

float networkUploadSpeed(const Connection *c);
float networkDownloadSpeed(const Connection *c);

int networkUpdateConnection(Connection *);
void networkKillConnection(Connection *);

#endif
