#include <allegro.h>
#include <string.h>
#include <assert.h>

#include "videodump.h"
#include "command.h"
#include "commandqueue.h"
#include "draw.h"
#include "config.h"
#include "arguments.h"

/* We're basically replacing client.c; and it had some functions that other files used.
   We'll replace them with dummies: */

uint8_t userid = 0;
int shadowColour(Colour c) {return 0;}
void chatAddMessage(const char *foo, const char *bar) {}
void userSetColour(uint8_t id, Colour c) {}
void userSetPos(uint8_t id, int16_t x, int16_t y) {}
const char * userGetName(uint8_t id) {return "";}

static void applyCommandHash(const Command *command, unsigned int *a, unsigned int *b)
 {
  const unsigned char *c = command->data_start, *e = command->data_end;
  
  *a += (unsigned int)command->type;
  *b += *a;
  *a %= 65521;
  *b %= 65521;
  
  for(; c != e; ++c)
   {
    *a += *c;
    *b += *a;
    *a %= 65521;
    *b %= 65521;
   }
 }

static const char * constructFileName(const char *name)
 {
  static char buf[256], *p;
  strncpy(buf, name, sizeof buf);
  p = buf+strlen(buf)-1;
  
  for(; p > buf && *p != '.'; --p)
   {}
  
  if(p <= buf)
   strncpy(buf, "out.bmp", sizeof buf);
  else
   strncpy(p, ".bmp", sizeof buf - (p - buf));
  
  return buf;
 }

typedef struct
 {
  BITMAP *buffer;
  const char *preview_file;
  unsigned int dump_w, dump_h;
  unsigned int delay;
  int clear;
  VideoDump *video;
 } Paint2BMP_State;
 
static int setPreviewFile(const char *arg, void *state)
 {
  ((Paint2BMP_State *)(state))->preview_file = arg;
  return 1;
 }

static int setVideoRez(const char *arg, void *_state)
 {
  Paint2BMP_State *state = (Paint2BMP_State *)_state;
  
  if(2 != sscanf(arg, "%ux%u", &state->dump_w, &state->dump_h))
   {
    fprintf(stderr, "Expected resolution in format '<width>x<height>'.\n");
    return 0;
   }
  
  if(state->dump_w == 0 || state->dump_h == 0)
   {
    fprintf(stderr, "Video width and height must be greater than zero.\n");
    return 0;
   }
  
  if((state->dump_w&1) || (state->dump_h&1))
   {
    fprintf(stderr, "Video width and height must be multiples of two.\n");
    return 0;
   }
  
  return 1;
 }

static int setVideoDelay(const char *arg, void *_state)
 {
  Paint2BMP_State *state = (Paint2BMP_State *)_state;
  
  if(1 != sscanf(arg, "%u", &state->delay))
   {
    fprintf(stderr, "Expected integer to represent delay as number of duplicated frames.\n");
    return 0;
   }
  
  return 1;
 }

static int clearImage(const char *arg, void *_state)
 {
  Paint2BMP_State *state = (Paint2BMP_State *)_state;
  
  if(0 == strcmp(arg, "y") || 0 == strcmp(arg, "yes") || 0 == strcmp(arg, "true") || 0 == strcmp(arg, "1"))
   {
    if(state->clear == 0)
     state->clear = 1;
   }
  else if(0 == strcmp(arg, "n") || 0 == strcmp(arg, "no") || 0 == strcmp(arg, "false") || 0 == strcmp(arg, "0"))
   {
    state->clear = 0;
   }
  else
   {
    fprintf(stderr, "Expected 'y' or 'n' for whether to clear at the start of each new image.\n");
    return 0;
   }
  
  return 1;
 }

static int dumpVideo(const char *arg, void *_state)
 {
  Paint2BMP_State *state = (Paint2BMP_State *)_state;
  
  if(state->video)
   {
    if(state->video->out != stdout)
     fclose(state->video->out);
    
    releaseVideoDump(state->video);
    state->video = 0;
   }
  
  if(!strcmp(arg, "-"))
   state->video = createVideoDump(stdout, state->buffer, state->dump_w, state->dump_h);
  else
   {
    FILE *file = fopen(arg, "wb");
    if(file && !(state->video = createVideoDump(file, state->buffer, state->dump_w, state->dump_h)))
     fclose(file);
   }
  
  if(!state->video)
   {
    fprintf(stderr, "Unable to dump video to %s.", arg);
    return 0;
   }
  
  return 1;
 }

static int doImage(const char *filename, void *_state)
 {
  Paint2BMP_State *state = (Paint2BMP_State *)_state;
  
  struct
   {
     unsigned int depth, hasha, hashb;
   } current, previous;
  
  bzf *file;
  unsigned int data = 0;
  const char *out = constructFileName(filename);
  Command command;
  
  current.depth = previous.depth = 0;
  current.hasha = previous.hasha = 1;
  current.hashb = previous.hashb = 1;
  
  if(!state->video && state->preview_file)
   {
    BITMAP *previous_bitmap;
    
    if((previous_bitmap = load_bitmap(out, 0)))
     {
      if(previous_bitmap->w != buffer_w || previous_bitmap->h != buffer_h)
       fprintf(stderr, "%s isn't the correct dimensions, rendering from scratch.\n", out);
      else
       {
        FILE *file = fopen(state->preview_file, "rb");
        
        if(file)
         {
          blit(previous_bitmap, state->buffer, 0, 0, 0, 0, buffer_w, buffer_h);
          
          if(1 != fread(&previous, sizeof previous, 1, file))
           {
            previous.depth = 0;
            previous.hasha = 1;
            previous.hashb = 0;
           }
          
          fclose(file);
         }
        else
         {
          previous.depth = 0;
          previous.hasha = 1;
          previous.hashb = 0;
         }
       }
      
      destroy_bitmap(previous_bitmap);
     }
    else fprintf(stderr, "%s doesn't exist, starting from scratch.\n", out);
   }
  
  retry:
  
  if(previous.depth == 0)
   {
    if(state->clear)
     {
      drawPaper(state->buffer);
      state->clear = 1;
     }
    
    current.depth = 1;
   }
  
  file = bz_OpenRead(filename);
  
  if(!file)
   {
    fprintf(stderr, "Error reading %s.\n", filename);
    return 0;
   }
  
  while(commandInitFromFile(&command, file))
   {
    if(state->video)
     {
      data += command.data_end - command.data_start;
      
      if(data > 128)
       {
        data -= 128;
        if(!performDump(state->video))
         {
          if(state->video->out != stdout)
           fclose(state->video->out);
          
          releaseVideoDump(state->video);
          
          state->video = 0;
          fprintf(stderr, "Error dumping video frame.\n");
          return 0;
         }
       }
     }
    
    if(current.depth >= previous.depth)
     {
      if(current.depth == previous.depth &&
         !(current.hasha == previous.hasha && current.hashb == previous.hashb))
       {
        fprintf(stderr, "Data doesn't belong to image. Will start over.\n");
        bz_CloseRead(file);
        current.depth = previous.depth = 0;
        current.hasha = previous.hasha = 1;
        current.hashb = previous.hashb = 1;
        goto retry;
       }
      
      commandDraw(state->buffer, &command);
     }
    
    ++current.depth;
    applyCommandHash(&command, &current.hasha, &current.hashb);
    
    commandUninit(&command);
   }
  
  bz_CloseRead(file);
  
  if(state->video)
   {
    unsigned int delay = 0;
    for(; delay < state->delay; ++delay)
     if(!(delay==0?performDump:duplicatePreviousFrame)(state->video))
      {
       if(state->video->out != stdout)
        fclose(state->video->out);
       
       releaseVideoDump(state->video);
       state->video = 0;
       fprintf(stderr, "Error dumping video frame.\n");
       return 0;
      }
   }
  
  if(current.depth < previous.depth)
   {
    fprintf(stderr, "Data doesn't belong to image. Will start over.\n");
    current.depth = previous.depth = 0;
    current.hasha = previous.hasha = 1;
    current.hashb = previous.hashb = 1;
    goto retry;
   }
  
  if(current.depth != previous.depth && (state->preview_file || !state->video))
   {
    if(save_bitmap(out, state->buffer, 0))
     {
      fprintf(stderr, "Error writing %s.\n", out);
      return 0;
     }
    else
     {
      fprintf(stderr, "Wrote %s.\n", out);
      
      if(state->preview_file)
       {
        FILE *file = fopen(state->preview_file, "wb");
        
        if(file)
         {
          fwrite(&current, 1, sizeof current, file);
          fclose(file);
         }
       }
     }
   }
  
  state->preview_file = 0;
  
  return 1;
 }

static int printHelp(const char *unused1, void *_state)
 {
  Paint2BMP_State *state = (Paint2BMP_State *)_state;
  
  fprintf(stderr,
          "Usage: paint2bmp [options] image [[options] image2 [ ... ]]]\n"
          "\n"
          "The order of the options is important! They must appear before the thing\n"
          "they're intended to affect."
          "\n");
  fprintf(stderr,
          " -p state_file  Save/Load the state of the next image to/from this file.\n"
          "                If the file is updated the drawing is continued rather\n"
          "                than generated from the beginning.\n");
  fprintf(stderr,
          " -r <w>x<h>     Select the resolution of the next video file.\n"
          "                Width and height must be positive multiples of two.\n"
          "                Defaults to 320x240. The maximum size is %ux%u.\n", state->buffer->w, state->buffer->h);
  fprintf(stderr,
          " -d <n>         Append n duplicate video frames to the end of the following\n"
          "                images. Defaults to 100.\n");
  fprintf(stderr,
          " -c [y|n]       Clear image between images. If 'n', the next image will be\n"
          "                drawn over the previous image. Defaults to 'y'.\n");
  fprintf(stderr,
          " -v video_file  Stop dumping bitmaps, and begin dumping images as a raw\n"
          "                YUV420p video stream. Example:\n"
          "\n"
          "                    paint2bmp -v - image | \\\n"
          "                    ffmpeg -f rawvideo -s 320x240 -i - video.avi\n"
          "\n");
  fprintf(stderr,
          " --help         Print this message.\n");
  
  return 1;
 }

static Argument p2b[] =
 {
  {"preview", 'p', "", RequireArgument, &setPreviewFile},
  {"resolution", 'r', "", RequireArgument, &setVideoRez},
  {"delay", 'd', "", RequireArgument, &setVideoDelay},
  {"video", 'v', "", RequireArgument, &dumpVideo},
  {"clear", 'c', "", RequireArgument, &clearImage},
  {0, 0, "", RequireArgument, &doImage},
  {"help", '?', "", NoArgument, &printHelp}
 };

int main(int argc, const char *argv[])
 {
  int result;
  Paint2BMP_State state;
  
  state.preview_file = 0;
  state.dump_w = 320;
  state.dump_h = 240;
  state.buffer = 0;
  state.delay = 100;
  state.video = 0;
  state.clear = 2; // 2 means the buffer is full of garbage, and needs to be cleared. 1 means clear at end of image. 0 means don't clear.
  
  install_allegro(SYSTEM_NONE, &errno, &atexit);
  set_color_depth(32);
  set_color_conversion(COLORCONV_NONE);
  
  if(!(state.buffer = create_bitmap_ex(32, buffer_w, buffer_h)))
   {
    fprintf(stderr, "Error creating buffer.\n");
    return 1;
   }
  
  result = parseArguments(argc, argv, p2b, p2b+sizeof(p2b)/sizeof(*p2b), &state);
  
  if(state.video)
   {
    if(state.video->out != stdout)
     fclose(state.video->out);
    
    releaseVideoDump(state.video);
   }
  
  destroy_bitmap(state.buffer);
  
  return result ? 0 : 1;
 } END_OF_MAIN()
