#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> 
#include <time.h>

#ifdef _WIN32
#include "windows.h"
#endif

#include "mem.h"
#include "network.h"
#include "command.h"
#include "commandqueue.h"
#include "version.h"
#include "user.h"
#include "log.h"

CommandQueue instructions;
Command command, repeat;

char motd[512];

void saveMOTD(void)
 {
  FILE *file = fopen("motd.txt", "wb");
  
  if(file)
   {
    fwrite(motd, strlen(motd), 1, file);
    fclose(file);
   }
 }

void loadMOTD(void)
 {
  FILE *file = fopen("motd.txt", "rb");
  
  if(file)
   {
    char *p = motd;
    int ch;
    
    while(EOF != (ch = fgetc(file)))
     {
      *p = ch;
      
      if(++p - motd == sizeof motd - 1)
       break;
     }
    
    *p = 0;
    fclose(file);
   }
  else
   *motd = 0;
 }

static const char preview_filename[] = "preview.polypaint";
static bzf *preview_file = 0;

static void closePreviewFile()
 {
  if(preview_file)
   {
    bz_CloseWrite(preview_file);
    preview_file = 0;
   }
 }

static void openPreviewFile()
 {
  closePreviewFile();
  
  preview_file = bz_OpenWrite(preview_filename);
 }

static void appendCommandToPreviewFile(Command *command)
 {
  if(preview_file && (!commandSave(command, preview_file) || !bz_Flush(preview_file)))
   closePreviewFile();
 }

uint8_t votes_restart = 0;

void serverDisconnectAll(void)
 {
  uint16_t i = 0;
  logMessage(Debug, "Disconnecting everyone.");
  
  for(; i < networkClients(); ++i)
   networkClient(i)->complete = 1;
 }

typedef struct
 {
  uint8_t mode, id, vote_kick, vote_restart, votes_against, kicks;
 } ClientInfo;

int serverUsers(void)
 {
  int i, users = 0;
  Connection *client;
  
  for(i = 0; i < networkClients(); ++i)
   if((client = networkClient(i)) && !client->complete && client->custom_data && ((ClientInfo *)(client->custom_data))->id)
    ++users;
  
  return users;
 }

void sendClientsMessage(const char *string)
 {
  int i;
  Connection *client;
  Command command;
  
  commandInit(&command, Chat);
  
  if(!commandPutU8(&command, 0) || !commandPutString(&command, string))
   {
    commandUninit(&command);
    return;
   }
  
  for(i = 0; i < networkClients(); ++i)
   if((client = networkClient(i)) && !client->complete && client->custom_data && ((ClientInfo *)(client->custom_data))->id)
    if(!commandNetworkSend(&command, client))
     client->complete = 1;
  
  commandUninit(&command);
 }

int sendClientMessage(Connection *client, const char *string)
 {
  Command command;
  
  commandInit(&command, Chat);
  
  if(!commandPutU8(&command, 0) || !commandPutString(&command, string))
   {
    commandUninit(&command);
    return 0;
   }
  
  if(!commandNetworkSend(&command, client))
   {
    client->complete = 1;
    return 0;
   }
  
  commandUninit(&command);
  return 1;
 }

ClientInfo * findClient(uint8_t id)
 {
  uint16_t i = 0;
  Connection *client;
  
  if(userValid(id))
   for(; i < networkClients(); ++i)
    if((client = networkClient(i)) && !client->complete && client->custom_data && ((ClientInfo *)(client->custom_data))->id == id)
     return (ClientInfo *)(client->custom_data);
  
  return 0;
 }

int ipGetKicks(const char *address)
 {
  FILE *file = fopen(address, "rb");
  
  if(file)
   {
    int kicks;
    
    if(!fread(&kicks, sizeof kicks, 1, file))
     kicks = 0;
    
    fclose(file);
    return kicks;
   }
  else return 0;
 }

void ipSetKicks(const char *address, int kicks)
 {
  FILE *file = fopen(address, "wb");
  
  if(file)
   {
    fwrite(&kicks, sizeof kicks, 1, file);
    fclose(file);
   }
 }

int serverConnect(Connection *c, void **_data)
 {
  int i = 0;
  unsigned int count = 0;
  ClientInfo *info;
  
  for(; i < networkClients(); ++i)
   if(networkClient(i) != c && !strcmp(networkClient(i)->address, c->address))
    ++count;
  
  if(count > 4)
   return 0; /* You surely don't need to log in more than four times. */
  
  if(!(info = (ClientInfo *)(*_data = malloc(sizeof(ClientInfo)))))
   return 0;
  
  info->mode = info->id = info->vote_kick = info->vote_restart = info->votes_against = 0;
  info->kicks = ipGetKicks(c->address);
  
  if(info->kicks >= 3) /* They're out of here! */
   return 0;
  
  return 1;
 }

int serverUpdate(Connection *c, void **_data)
 {
  ClientInfo *info = (ClientInfo *)(*_data);
  
  if(!info)
   return 0;
  
  if(c->recv_end - c->recv_start > 4096)
   {
    ++info->kicks;
    return 0; /* flooding us with crap?! */
   }
  
  if(commandInitFromNetwork(&command, c))
   {
    uint16_t i;
    Connection *client;
    
    if(command.data_end - command.data_start > 1300)
     { /* Command is too large; probably not valid. */
      ++info->kicks;
      commandUninit(&command);
      return 0;
     }
    
    switch(command.type)
     {
      case Login: /* Trying to login. . . */
       
       if(info->mode != 0) /* They're already logged in! */
        {
         ++info->kicks;
         commandUninit(&command);
         return 0;
        }
       else
        {
         char username[16];
         uint32_t version = commandGetU32(&command);
         
         commandGetString(&command, username, sizeof username);
         commandUninit(&command);
         
         if(version != VERSION) /* They're using an incompatible client version. */
          {
           char buf[128];
           snprintf(buf, sizeof buf, "Warning: Version miss match. Client version: %d.%d; Server version: %d.%d. Make sure you have the latest version!", VER_GET_MAJOR(version), VER_GET_MINOR(version), VERSION_MAJOR, VERSION_MINOR);
           if(!sendClientMessage(c, buf))
            return 0;
          }
         
         if(*motd)
          {
           char buf[512];
           snprintf(buf, sizeof buf, "MOTD: %s", motd);
           
           if(!sendClientMessage(c, buf))
            return 0;
          }
         
         if(!(info->id = userGenerateNew(username))) /* Get them a new user id. */
          return 0;
         
         commandInit(&command, Login);
         
         if(!commandPutU8(&command, info->id) || !commandPutString(&command, userGetName(info->id)))
          { /* Error storing their id and name in the login command. */
           userForget(info->id);
           commandUninit(&command);
           return 0;
          }
         
         for(i = 0; i < networkClients(); ++i) /* Add all the users to the login command to so they know who's logged in. */
          if((client = networkClient(i)) != c && !client->complete && ((ClientInfo *)(client->custom_data))->mode == 1)
           if(!commandPutU8(&command, ((ClientInfo *)(client->custom_data))->id) || !commandPutString(&command, userGetName(((ClientInfo *)(client->custom_data))->id)))
            { /* Something went fubar. . . */
             userForget(info->id);
             commandUninit(&command);
             return 0;
            }
         
         if(!commandNetworkSend(&command, c))
          { /* Couldn't send them their login command. */
           userForget(info->id);
           commandUninit(&command);
           return 0;
          }
         
         commandUninit(&command);
         
         for(i = 0; i < commandQueueElements(&instructions); ++i) /* Send them all the instructions we've recieved so far. */
          if(!commandNetworkSend(commandQueueElement(&instructions, i), c))
           {
            userForget(info->id);
            return 0;
           }
         
         if(!commandNetworkSend(&repeat, c))
          {
           userForget(info->id);
           return 0; /* Couldn't send the repeat command to tell them we're done with their login request. */
          }
         
         commandInit(&command, Join);
         
         if(!commandPutU8(&command, info->id) || !commandPutString(&command, userGetName(info->id)))
          { /* Couldn't add the user id and name to the join command. */
           userForget(info->id);
           commandUninit(&command);
           return 0;
          }
         
         /* From this point on they're considered offically logged in. Errors become the other clients problems. */
         logMessage(Debug, "User %d (%s) joining.", info->id, userGetName(info->id));
         info->mode = 1;
         
         for(i = 0; i < networkClients(); ++i) /* Send the login command to all the connected users. */
          if((client = networkClient(i)) != c && !client->complete && client->custom_data && ((ClientInfo *)(client->custom_data))->mode == 1)
           if(!commandNetworkSend(&command, client))
            client->complete = 1;
         
         if(info->kicks)
          {
           char msg[128];
           snprintf(msg, sizeof msg, "%s has %d chance%s to behave before an IP ban.", username, 3 - info->kicks, ((3 - info->kicks) == 1) ? "" : "s");
           sendClientsMessage(msg);
          }
         
         commandUninit(&command);
         return 1;
        }
      
      case Pen:
      case Brush:
      case Pencil:
      case Chalk:
      case Smudge:
      case Stamp:
      case Chat:
       
       if(info->mode == 0)
        { /* Trying to draw and they're not yet logged in. */
         ++info->kicks;
         commandUninit(&command);
         return 0;
        }
       
       if(commandGetU8(&command) != info->id)
        { /* They're trying to draw as someone they're not. */
         ++info->kicks;
         commandUninit(&command);
         return 0;
        }
       
       logMessage(Debug, "New drawing related command (%d) from user %d (%s).", command.type, info->id, userGetName(info->id));
       
      break;
      
      case Restart:
       {
        uint8_t state = commandGetU8(&command)?1:0;
        
        if(info->vote_restart != state)
         {
          char buf[128];
          votes_restart += (state ? 1 : -1);
          info->vote_restart = state;
          logMessage(Debug, "Restart vote is %d%%.", votes_restart*100/serverUsers());
          snprintf(buf, sizeof buf, "Restart vote is %d%%.", votes_restart*100/serverUsers());
          sendClientsMessage(buf);
         }
        
        commandUninit(&command);
        return 1;
       }
      
      case Kick:
       {
        uint8_t state = commandGetU8(&command);
        
        if(info->vote_kick != state)
         {
          char buf[128];
          ClientInfo *data;
          
          if((data = findClient(info->vote_kick)))
           {
            --data->votes_against;
            snprintf(buf, sizeof buf, "%s's warning level is %d%%.", userGetName(data->id), data->votes_against*100/serverUsers());
            sendClientsMessage(buf);
           }
          
          if((data = findClient(state)) && data->id != info->id)
           {
            ++data->votes_against;
            snprintf(buf, sizeof buf, "%s's warning level is %d%%.", userGetName(data->id), data->votes_against*100/serverUsers());
            sendClientsMessage(buf);
            info->vote_kick = state;
           }
          else info->vote_kick = 0;
         }
        
        commandUninit(&command);
        return 1;
       }
      
      case Motd:
       {
        char pass[32];
        char buf[512];
        commandGetString(&command, pass, sizeof pass);
        
        if(strcmp(pass, "rosebud")) /* Todo: Change the password. */
         {
          sendClientMessage(c, "Bad password.");
          commandUninit(&command);
          return 1;
         }
        
        commandGetString(&command, motd, sizeof motd);
        saveMOTD();
        
        if(*motd)
         snprintf(buf, sizeof buf, "%s set MOTD to: %s", userGetName(info->id), motd);
        else
         snprintf(buf, sizeof buf, "MOTD removed by %s.", userGetName(info->id));
        
        sendClientsMessage(buf);
        return 1;
       }
      
      case RepeatCommand:
      case Join:
      case Leave:
      default: /* Clients shouldn't send these commands. */
       commandUninit(&command);
       return 0;
     }
    
    if(!commandNetworkSend(&repeat, c))
     { /* Couldn't let them know we got their command and accepted it. */
      commandUninit(&command);
      return 0;
     }
    
    for(i = 0; i < networkClients(); ++i) /* Relay the command to all the other clients. */
     if((client = networkClient(i)) != c && !client->complete)
      if(!commandNetworkSend(&command, client))
       client->complete = 1;
    
    if(command.type == Chat) /* Don't record chatting. */
     {
      commandUninit(&command);
      return 1;
     }
    
    appendCommandToPreviewFile(&command);
    
    if(!commandQueuePushBack(&instructions, &command))
     { /* Couldn't store the command. We've obviously can't
          continue on like this; disconnect everyone and start over. */
      commandUninit(&command);
      serverDisconnectAll();
      
      /* Remove all the instructions. */
      while(commandQueuePopBack(&instructions, &command))
       commandUninit(&command);
      
      openPreviewFile();
      
      return 0;
     }
    
    logMessage(Debug, "%d commands recorded to send to new clients.", commandQueueElements(&instructions));
   }
  
  return 1;
 }

void serverDisconnect(Connection *c, void **_data)
 {
  ClientInfo *info = (ClientInfo *)(*_data);
  
  if(info && info->mode == 1)
   {
    uint16_t i;
    Connection *client;
    
    info->mode = 2;
    
    commandInit(&command, Leave);
    
    logMessage(Debug, "User %d (%s) leaving.", info->id, userGetName(info->id));
    ipSetKicks(c->address, info->kicks);
    
    commandQueueForgetUser(&instructions, info->id);
    userForget(info->id);
    
    if(!commandPutU8(&command, info->id))
     {
      commandUninit(&command);
      serverDisconnectAll();
      return;
     }
    
    if(info->vote_restart)
     --votes_restart;
    
    for(i = 0; i < networkClients(); ++i)
     if((client = networkClient(i)) && !client->complete && client->custom_data && ((ClientInfo *)(client->custom_data))->mode == 1)
      {
       if(((ClientInfo *)(client->custom_data))->vote_kick == info->id)
        ((ClientInfo *)(client->custom_data))->vote_kick = 0;
       
       if(info->vote_kick == ((ClientInfo *)(client->custom_data))->id)
        --((ClientInfo *)(client->custom_data))->votes_against;
       
       if(!commandNetworkSend(&command, client))
        client->complete = 1;
      }
    
    commandUninit(&command);
   }
 }

const char * createImageFilename(void)
 {
  static char buf[256];
  
  time_t date = time(0);
  strftime(buf, sizeof buf, "./save/%b-%d-%Y.%H-%M-%S.polypaint", localtime(&date));
  
  return buf;
 }

int main(int argc, char *argv[])
 {
  #ifndef _WIN32
  static const struct timespec sleep_time = { 0, 20000000 };
  #endif
  static char cwd_buf[512];
  const char *wd = 0;
  time_t birth = time(0), last_active = time(0);
  
  for(--argc, ++argv; argc; --argc, ++argv)
   {
    if(chdir(*argv))
     {
      logMessage(Error, "Unable to set working directory to: %s", *argv);
      return EXIT_FAILURE;
     }
    else wd = *argv;
   }
  
  if(!wd)
   {
    if(cwd_buf != getcwd(cwd_buf, sizeof cwd_buf))
     {
      logMessage(Error, "Unable to get working directory name.");
      return EXIT_FAILURE;
     }
    
    wd = cwd_buf;
   }
  
  logMessage(Notice, "Server started in: %s", wd);
  
  loadMOTD();
  
  commandInit(&repeat, RepeatCommand);
  
  if(!commandQueueInitFilename(&instructions, preview_filename) && !commandQueueInit(&instructions))
   {
    commandUninit(&repeat);
    return EXIT_FAILURE;
   }
  
  if(!networkInit())
   {
    commandQueueUninit(&instructions);
    commandUninit(&repeat);
    return EXIT_FAILURE;
   }
  
  if(!networkServerInit(2020))
   {
    logMessage(Error, "Unable to start server.");
    networkUninit();
    commandQueueUninit(&instructions);
    commandUninit(&repeat);
    return EXIT_FAILURE;
   }
  
  openPreviewFile();
  
  if(preview_file && commandQueueElements(&instructions))
   { /* Now that we're loading the old image, need to resave the preview of it. */
    commandQueueSave(&instructions, preview_file);
    bz_Flush(preview_file);
   }
  
  while(1)
   {
    uint16_t users;
    uint16_t i;
    Connection *client;
    float age = time(0) - birth;
    if(age < 10) age = 10;
    
    #ifdef _WIN32
    Sleep(20);
    #else
    nanosleep(&sleep_time, 0);
    #endif
    
    networkUpdateServer(&serverConnect, &serverUpdate, &serverDisconnect);
    
    if((users = serverUsers()))
     {
      if(votes_restart*100/users > 50)
       { /* Restarting! */
        if(commandQueueElements(&instructions))
         {
          openPreviewFile();
          
          if(commandQueueElements(&instructions) > 20)
           commandQueueSaveFilename(&instructions, createImageFilename());
          
          while(commandQueuePopBack(&instructions, &command))
           commandUninit(&command);
          
          for(i = 0; i < networkClients(); ++i)
           if((client = networkClient(i)) && !client->complete && client->custom_data && ((ClientInfo *)(client->custom_data))->mode == 1)
            if(!(commandNetworkSend(&command, client)))
             client->complete = 1;
          
          commandUninit(&command);
         }
        
        for(i = 0; i < networkClients(); ++i)
         if((client = networkClient(i)) && !client->complete && client->custom_data && ((ClientInfo *)(client->custom_data))->mode == 1)
          ((ClientInfo *)(client->custom_data))->vote_restart = 0;
        
        votes_restart = 0;
        
        continue;
       }
       
      for(i = 0; i < networkClients(); ++i)
       if((client = networkClient(i)) && !client->complete && client->custom_data && ((ClientInfo *)(client->custom_data))->votes_against*100/users > 50)
        {
         ++((ClientInfo *)(client->custom_data))->kicks;
         client->complete = 1;
        }
      
      last_active = time(0);
     }
    else if(commandQueueElements(&instructions) &&
            (time(0)-last_active) > 300)
     { /* If nobody has been connected in the last 5 minutes, throw out the current image. */
      if(commandQueueElements(&instructions) > 20)
       commandQueueSaveFilename(&instructions, createImageFilename());
      
      while(commandQueuePopBack(&instructions, &command))
       commandUninit(&command);
      
      openPreviewFile();
     }
     
    #ifndef NDEBUG
    printf("\n\n\n --- %d Clients Connected ---\n", networkClients());
    
    uint32_t up = 0, down = 0;
    for(i = 0; i < networkClients(); ++i)
     if((client = networkClient(i)) && !client->complete)
      {
       printf(" %s: %s\tUp: %dB (%.2fKB/S)\tDown: %dB (%.2fKB/S)\n", userGetName(((ClientInfo *)(client->custom_data))->id), client->address, client->sent, networkUploadSpeed(client)/1024.0, client->recieved, networkDownloadSpeed(client)/1024.0);
       up += client->sent;
       down += client->recieved;
      }
    
    printf(" *: *.*.*.*\tUp: %dB (%.2fKB/S)\tDown: %dB (%.2fKB/S)\n", up, (float)up/age/1024.0, down, (float)down/age/1024.0);
    #endif
   }
  
  /* Can't be reached:
  
  commandUninit(&repeat);
  networkServerUninit();
  networkUninit();
  commandQueueUninit(&instructions);
  return EXIT_SUCCESS;
  */
 }

